#include <OS.h>

#include <string.h>

#include <system/MessageTransceiverThread.h>

#include "GameSession.h"

enum {
	HELLO = 'gshl',
	HI_REJECT = 'gshr',
	HI_ACCEPT = 'gsha',
	GOOD_BYE = 'gsgb'
};

using namespace muscle;

class GameWorkerSession : public ThreadWorkerSession {
	public:
		GameWorkerSession(GameSession *);
		
		virtual void AsyncConnectCompleted();
		virtual bool ClientConnectionClosed();
		virtual void MessageReceivedFromGateway(MessageRef msg);

	protected:
		virtual const char * GetTypeName() const;
		
	private:
		friend class GameSession;
		GameSession *session;
		
		bool connected;
};

GameSession::GameSession(const char *server, int port, const char *user_name) :
  Thread(),
  return_uid(-1),
  connected_to(-1),
  mtt(new MessageTransceiverThread()),
  worker(new GameWorkerSession(this)),
  connect_thread_waiting(0),
  connect_result(false),
  sn(user_name) {
  	StartInternalThread();
	mtt->StartInternalThread();
	mtt->AddNewConnectSession(server,port,AbstractReflectSessionRef(worker,NULL,false));
}

GameSession::~GameSession() {
	if (connected_to >= 0)
		EndGame();
	ShutdownInternalThread();
	return_uid = connected_to = -1;
	connect_result = false;
	if (connect_thread_waiting > 0)
		resume_thread(connect_thread_waiting);
	
	mtt->ShutdownInternalThread();
	delete mtt;
}

void GameSession::MessageReceivedFromPartner(muscle::MessageRef msg) {}
void GameSession::MessageReceivedFromNetwork(muscle::MessageRef msg) {}

status_t GameSession::MessageReceivedFromOwner(MessageRef msg, uint32) {
	if (msg() == NULL) //---Why on earth does this happen?
		return B_ERROR;
	
	switch (msg()->what) {
		case 'gscn':
			if (!worker->connected) {
				worker->connected = true;
				ServerConnectDone();
				break;
			} else {
				goto who_knows;
			}
		case 'gsdc':
			if (!worker->connected) {
				worker->connected = true;
				ServerConnectDone();
				break;
			} else {
				goto who_knows;
			}
		case HELLO: {
			const char *uname;
			int32 foreign_uid;
			msg()->FindString("user_name",&uname);
			msg()->FindInt32("return_uid",&foreign_uid);
			MessageRef response = GetMessageFromPool(AllowConnectFrom(uname,foreign_uid) ? HI_ACCEPT : HI_REJECT);
			//response()->AddMessage("original",msg);
			response()->AddInt32("return_uid",return_uid);
			String string;
			string << "/*/" << int(foreign_uid);
			response()->AddString(PR_NAME_KEYS,string);
			if (response()->what == HI_ACCEPT)
				connected_to = foreign_uid; //---Implicit HI_ACCEPT
			SendMessageToServer(response);
			break; }
		case HI_REJECT:
			connect_result = false;
			resume_thread(connect_thread_waiting);
			connect_thread_waiting = 0;
			break;
		case HI_ACCEPT:
			connect_result = true;
			msg()->FindInt32("return_uid",&connected_to);
			resume_thread(connect_thread_waiting);
			connect_thread_waiting = 0;
			break;
		case GOOD_BYE:
			PartnerGone();
			break;
		case PR_RESULT_DATAITEMS: {
			String name;
			type_code bs;
			int32 id;
			const char *uname;
			MessageFieldNameIterator foo = msg()->GetFieldNameIterator();
			type_code type;
			while (foo.GetNextFieldName(name) == B_OK) {
				msg()->GetInfo(name,&type);
				if ((name == PR_NAME_REMOVED_DATAITEMS) || (type != B_MESSAGE_TYPE))
					continue;
				parse_out_uid_uname(name(),&id,&uname);
				NewUserLoggedOn(id,uname);
			}
			for (int32 i = 0; msg()->FindString(PR_NAME_REMOVED_DATAITEMS,i,name) == B_OK; i++) {
				parse_out_uid_uname(name(),&id,&uname);
				if (id == connected_to)
					PartnerGone();
				UserLoggedOff(id,uname);
			}
			break; }			
		case PR_RESULT_PARAMETERS: {
			const char * sessionRoot;
			if (msg()->FindString(PR_NAME_SESSION_ROOT, &sessionRoot) == B_NO_ERROR) {
				// session root is of form "/hostname/sessionID"; parse these out
				const char * lastSlash = strrchr(sessionRoot, '/');
				if (lastSlash)
					sscanf(lastSlash+1,"%d",&return_uid);
				else
					puts("Funky shit is going on with this MUSCLE server!");
			}
			break; }
		default:
		who_knows:
			if (connected_to == -1)
				MessageReceivedFromNetwork(msg);
			else {
				int32 from;
				msg()->FindInt32("return_uid",&from);
				if (from == connected_to)
					MessageReceivedFromPartner(msg);
				else
					MessageReceivedFromNetwork(msg);
			}
			break;
	}
	
	return B_OK;
}

void GameSession::parse_out_uid_uname(const char *key, int32 *uid, const char **uname) {
	//----/hostname/sessionID/GameSession/name_username
	char *a,*b = strdup(key);
	char *foob = b;
	volatile bool kill_parse_loop = false;
	for (int32 i = 0; (!kill_parse_loop) && ((a=b) != NULL);i++) {
		b = strchr(a,'/');
		if (b != NULL) {
			*b = 0; b++;
		}
		switch(i) {
			case 2: //---sessionID
				sscanf(a,"%d",uid);
				break;
			case 3: //---check for gs
				if (strcmp(a,GetTypeName()) != 0)
					kill_parse_loop = true;
				break;
			case 4:
				if (strncmp(a,"name_",5) != 0) {
					kill_parse_loop = true;
					break;
				}
				a += 5;
				kill_parse_loop = true; //--We'll just pass a
				break;
		}
	}
	*uname = (a-foob) + key;
	free(foob);
	
}
	

bool GameSession::AllowConnectFrom(const char *user_name, int32 id) {
	if (connected_to < 0)
		return true;
		
	return false;
}

bool GameSession::ConnectTo(const char *user) {
	String dest;
	dest << GetTypeName() << "/name_" << user;
	MessageRef message = GetMessageFromPool(HELLO);
	message()->AddString(PR_NAME_KEYS,dest);
	message()->AddString("user_name",sn);
	message()->AddInt32("return_uid",return_uid);
	SendMessageToServer(message);
	connect_thread_waiting = find_thread(NULL);
	suspend_thread(connect_thread_waiting);
	return connect_result;
}
	
bool GameSession::ConnectTo(int32 user_id) {
	String dest;
	dest << "/*/" << int(user_id);
	MessageRef message = GetMessageFromPool(HELLO);
	message()->AddString(PR_NAME_KEYS,dest);
	message()->AddString("user_name",sn);
	message()->AddInt32("return_uid",return_uid);
	SendMessageToServer(message);
	connect_thread_waiting = find_thread(NULL);
	suspend_thread(connect_thread_waiting);
	return connect_result;
}

void GameSession::ServerDisconnected() {
	PartnerGone();
	
	return_uid = connected_to = -1;
	connect_result = false;
	if (connect_thread_waiting > 0)
		resume_thread(connect_thread_waiting);
	
	mtt->ShutdownInternalThread();
	delete mtt;
	//-----Hrrm, we can't delete ourself, because that will kill our own thread
}

void GameSession::ServerConnectDone() {
	MessageRef message = GetMessageFromPool(PR_COMMAND_SETDATA);
	Message name;
	String foo;
	foo << GetTypeName() << "/name_" << sn;
	message()->AddMessage(foo,name);
	SendMessageToServer(message);
	
	//--------Get out user id
	message = GetMessageFromPool(PR_COMMAND_GETPARAMETERS);
	SendMessageToServer(message);
	
	//--------Set data subscription
	message = GetMessageFromPool(PR_COMMAND_SETPARAMETERS);
	foo = "";
	foo << "SUBSCRIBE:/*/*/" << GetTypeName() << "/name_*";
	message()->AddBool(foo,true);
	SendMessageToServer(message);
}

void GameSession::SendMessageToPartner(muscle::MessageRef ref) {
	ref()->AddInt32("return_uid",return_uid);
	ref()->AddString("user_name",sn);
	
	muscle::String foo;
	foo << "/*/" << int(connected_to);
	ref()->AddString(PR_NAME_KEYS,foo);
	SendMessageToServer(ref);
}

void GameSession::SendMessageToServer(muscle::MessageRef ref) {	
	mtt->SendMessageToSessions(ref);
}

void GameSession::PartnerGone() {
	connected_to = -1;
}

void GameSession::EndGame() {
	SendMessageToPartner(GetMessageFromPool(GOOD_BYE));
	PartnerGone();
}

void GameSession::ChangeName(const char *new_user_name) {
	//-----First, toss the old name entry
	MessageRef ref = GetMessageFromPool(PR_COMMAND_REMOVEDATA);
	String foo;
	foo << GetTypeName() << "/name_" << sn;
	ref()->AddString(PR_NAME_KEYS,foo);
	SendMessageToServer(ref);
	
	//-----Then, create a new one
	sn = new_user_name;
	ref = GetMessageFromPool(PR_COMMAND_SETDATA);
	Message name;
	foo = "";
	foo << GetTypeName() << "/name_" << sn;
	ref()->AddMessage(foo,name);
	SendMessageToServer(ref);
}

void GameSession::NewUserLoggedOn(int32, const char *) {}

void GameSession::UserLoggedOff(int32, const char *) {}

GameWorkerSession::GameWorkerSession(GameSession *foo) :
	ThreadWorkerSession(),
	session(foo),
	connected(false) {}

void GameWorkerSession::AsyncConnectCompleted() {
	session->SendMessageToInternalThread(GetMessageFromPool('gscn'));
	AbstractReflectSession::AsyncConnectCompleted();
}

bool GameWorkerSession::ClientConnectionClosed() {
	connected = false;
	session->SendMessageToInternalThread(GetMessageFromPool('gsdc'));
	return AbstractReflectSession::ClientConnectionClosed();; //-----Um, okay
}

void GameWorkerSession::MessageReceivedFromGateway(MessageRef msg) {
	session->SendMessageToInternalThread(msg);
	//AbstractReflectSession::MessageReceivedFromGateway(msg);
}

const char * GameWorkerSession::GetTypeName() const {
	return "GWS";
}

